/*
 * This file is part of the YASMIN 8051 simulator project
 *
 * Copyright (c) 2000 by Gwenn Englebienne
 *
 * This software is free software; it is protected by the GPL. You can 
 * modify and/or redistribute it under the terms specified by the GNU
 * Public Licence version 2.
 *
 * Be advised that this software comes with no warranty at all, not
 * even the implied warranty of merchantability or fitness for a 
 * particular purpose. 
 *
 * For more information, you could look at the COPYING file, or have a 
 * look at the free software foundation's web site: http://www.fsf.org
 *
 */
 
#include <iostream.h>
#include <iomanip.h>
#include <qfiledialog.h>
#include "8051.h"
#include "memory.h"
#include "general.h"
#include "linetext.h"

#define incLN if (text->getNumLines() > text->getCurLineNum () + 1) text->gotoNextLine()
#define DEBUG_2 0
#define CY 0x80
#define NOT_CY 0x7F
#define AC 0x40
#define NOT_AC 0xBF
#define OV 0x04
#define NOT_OV 0xFB
#define LABEL_LEN 20
#define INSTR_LEN 10
#define DEST_LEN 15
#define SOURCE_LEN 15
#define OP3_LEN 15                                                             

union convert {
     struct {
          char low, high;
     } split;
     unsigned short val;
};

Sim8051::Sim8051 (QWidget *parent, const char *name)
     : QGrid (4 ,parent, name)
{
     text = new DebugWidget (this);
     sb = new QScrollBar (0, text->getNumLines(),
                          1, text->getHeightInLines (), 0,
                          QScrollBar::Vertical, this, "Text scroll");
     connect (sb, SIGNAL (valueChanged (int)),
              text, SLOT (changeOrigin (int)));
     connect (text, SIGNAL (originChange (int)),
              this, SLOT (changeScroll (int)));
     results = new MemView (this);
     overview = new MemHex (this);
     err = new QMultiLineEdit (this);
     err->setMaximumSize (500,100);
     QLabel *junk = new QLabel ("", this);
     info = new ShowInfo (this);
     pv = new PortsView (this);
     setMargin (5);
     setSpacing (10);
     timer = new QTimer (this);
     connect (timer, SIGNAL (timeout ()), this, SLOT (step ()));
//     create_sfr_table();
     createMT();
     reset();
}


Sim8051::~Sim8051 ()
{
     delete text;
     delete results;
     delete overview;
     delete err;
     delete info;
     delete pv;
}


void Sim8051::fetch ()
{
     if (text->hasChanged())
          text->createLT (&labels);
}

void Sim8051::decode ()
{
     char *src = text->getCurLine ();
     if (!src) {
          label[0] = instr[0] = op1[0] = op2[0] = op3[0] = 0;
          return;
     }
     char *dst = label;
     while (*src && *src != ';' && *src != ':' && *src != ' ' && *src != '\t')
          *dst = *src, dst++, src++;
     *dst = 0;
     while (*src && (*src == ':' || *src == ' ' || *src == '\t')) src++;
     
     dst = instr;
     while (*src && *src != ';' && *src != ' ' && *src != '\t')
          *dst = *src, dst++, src++;
     *dst = 0;
     while (*src && (*src == ' ' || *src == '\t')) src++;
     
     dst = op1;
     while (*src && *src != ';' && *src != ',' && *src != ' ' && *src != '\t')
          *dst = *src, dst++, src++;
     *dst = 0;
     while (*src && (*src == ',' || *src == ' ' || *src == '\t')) src++;
     
     dst = op2;
     while (*src && *src != ';' && *src != ',' && *src != ' ' && *src != '\t')
          *dst = *src, dst++, src++;
     *dst = 0;
     while (*src && (*src == ',' || *src == ' ' || *src == '\t')) src++;
     
     dst = op3;
     while (*src && *src != ';' && *src != ' ' && *src != '\t')
          *dst = *src, dst++, src++;
     *dst = 0;
     uppercase (label);
     uppercase (instr);
     uppercase (op1);
     uppercase (op2);
     uppercase (op3);
}

void Sim8051::execute ()
{
     while (!*instr && !*op1) {
          incLN;
          decode ();
     }
     if (mnemonic.contains (instr))
          if (!(da (PCON) & 0x03)) // PD or IDL
               (this->*mnemonic.get(instr))();
          else; //sleep...
     else if (!strcmp (op1, "EQU")) {
          SFR tmp; // psoedo SFR with new variable name
          if (varNameOk (instr))
               if (is_a_nr (op2)) {
                    cerr << "Adding " << instr << " to sfr_info as " << op2 <<
                         endl;
                    tmp.type = byte;
                    tmp.address = asm_to_char (op2);
                    sfr_info.add (instr, tmp);
               }
               else
                    error ("This is not a valid address");
          else
               error ("This is not a valid name");
          incLN;
     }
     else {
          error ("This mnemonic is not valid.");
          incLN;
     }
     set_parity();
     decode ();
     while (!*instr && !*op1) {
          incLN;
          decode ();
     }
}

void Sim8051::step ()
{
     fetch();
     decode ();
     execute ();
     if (text->isBreak (text->getCurLineNum ()))
          stop ();
     check_system();
     if (!noOutput) {
          results->update();
          overview->update();
          info->setVal (cycles);
     }
     pv->update ();
     if (specialStep && level == stopLevel)
          stop ();
}

void Sim8051::check ()
{
     unsigned int here, old = text->getCurLineNum();
     for (here = 0; here < text->getNumLines (); here++) {
          text->setCurLine (here);
          decode();
          execute();
     }
     reset();
}

void Sim8051::changeScroll (int y)
{
     sb->setRange (0, text->getNumLines () - text->getHeightInLines ());
     sb->setSteps (1, text->getHeightInLines());
     sb->setValue (y);
}

int Sim8051::varNameOk (char *name) // assumes that name is uppercase
{
     char *tmp = name;
     if (*tmp >= 'A' && *tmp <= 'Z') { // first character is a letter
          *tmp++;
          while (*tmp && ((*tmp >= 'A' && *tmp <= 'Z')
                          || (*tmp >= '0' && *tmp <= '9')
                          || *tmp == '_'))
               tmp++;
     }
     return !*tmp; // if end of string was reached, return 1, else return 0
}

void Sim8051::run ()
{
     noOutput = 0;
     timer->start(100);
}

void Sim8051::runFast ()
{
     noOutput = 1;
     timer->start(1);
}

void Sim8051::stop ()
{
     specialStep = 0;
     timer->stop();
     noOutput = 0;
     results->update();
     overview->update();
     info->setVal (cycles);
}

void Sim8051::newFile ()
{
     reset ();
     text->newFile ();
}

void Sim8051::load ()
{
     reset();
     text->loadFile ();
}

void Sim8051::save ()
{
     text->saveFile ();
}

void Sim8051::saveAs ()
{
     QString tmp = QFileDialog::getSaveFileName("", "*.asm");
     if (!tmp.isNull()) 
          text->saveFile (tmp.ascii());
}

void Sim8051::stepIn ()
{
     stop ();
     step ();
}

void Sim8051::stepOut ()
{
     specialStep = 1;
     stopLevel = level - 1;
     runFast();
}

void Sim8051::stepOver ()
{
     specialStep = 1;
     stopLevel = level;
     runFast ();
}

void Sim8051::error (const char* s)
{
     char buf[150];
     snprintf (buf, 150, "Error (line %d): %s", text->getCurLineNum (), s);
     err->append (buf);
     stop();
}                                                                              

char Sim8051::source_val (void)
{
     char source_value;
     if (is_a_nr(op2))
          source_value = da (asm_to_char(op2));
     else if (op2[0] == '#')
         if (is_a_nr (op2 + 1))
	     source_value = asm_to_char (op2 + 1);
	 else
	     error ("Source operator is a bad constant.");
     else if (op2[0] == '@') {
          if (op2[1] == 'R' && (op2[2] == '0' ||
                                        op2[2] == '1') &&
              op2[3] == 0) {
               source_value = ida ( da (sfr_info[op2 +1 ].address));
          }
          else
               error ("Indirect works only with R0 & R1");
     }
     else if (sfr_info.contains(op2))
          source_value = da (sfr_info[op2].address);
     else
          error ("Invalid source");
     return source_value;
}

void Sim8051::unknown(void)
{
     char buf[50];
     snprintf (buf, 50, "The mnemonic [%s] hasn't been implemented yet.");
     error (buf);
     stop();
     incLN;
}

void Sim8051::no_movc(void)
{
     error("Because of the internal construction of this simulator,");
     error("the MOVC instruction isn't, and will probably never be implemented/");
     error("Any flames will be happilly sent to /dev/null. :->");
     incLN;
}

void Sim8051::end (void)
{
     // stop the timer
     stop();
}

void Sim8051::jump (void)
{
     if (labels.contains(op1)) {
          if (DEBUG_2)
               cout << "Jumping to " << op1 << ", line "
                    << labels[op1] << "\n";
          text->setCurLine (labels[op1]);
     }
     else if (strcmp ("$", op1))
     {
          error ("Unknown label.");
          incLN;
     }
     cycles += 2;
}

#define BIT 1
#define BYTE 0
#define WORD 456

void Sim8051::move (void)
{
     unsigned char dest_addr;
     char source_value;
     int direct = 1, source_type = 0, dest_type = 0;
     
     if (!strcmp(op1, "C")) {
          dest_type = BIT;
          dest_addr = sfr_info[op1].address;
     }
     else if (is_a_nr(op1)) {
          dest_addr = asm_to_char (op1);
          dest_type = BYTE;
          cycles++; // dest = direct ==> 2 clock cycles.
     }
     else if (op1[0] == '@') {   // it's indirect
          if (op1[1] == 'R' &&
              (op1[2] == '0' || op1[2] == '1') &&
              op1[3] == 0) {                      // Allow R0 and R1, nothing else
               if (DEBUG_2)
                    cout << "Indirect dest addr: "
                         << (int) sfr_info[op2].address
                         << ", addr2: "
                         << (int) da (sfr_info[op2].address)
                         << ", value = "
                         << (int) ida (da (sfr_info[op2].address))
                         << endl;
               dest_addr = da(sfr_info[op1 + 1].address);
               dest_type = BYTE;
               direct = 0;             // like I said, it's indirect
          }
          else
               error("Indirect move only works with R0 and R1");
     } else if (sfr_info.contains(op1))
          if (sfr_info[op1].type == byte) {
               if (DEBUG)
                    cerr << "dest_addr = "
                         << (int)sfr_info[op1].address << endl;
               dest_addr = sfr_info[op1].address;
               dest_type = BYTE;
          }
          else {
               dest_type = BIT;
               dest_addr = sfr_info[op1].address;
          }
     else if (!strcmp (op1, "DPTR"))  // It's a word, so let's avoid trouble: do it all here...
          if (op2[0] == '#') {
               set_dptr (asm_to_int (op2+1));
               incLN;
               return;                      // ... and bail out
          }
	  else;
     else
          error ("Invalid destination");
     
     if (!strcmp(op2, "C")) {   // move from carry
          source_type = BIT;
          source_value = is_bit_set (sfr_info["C"].address);
     }
     else if (is_a_nr(op2)) {    // move from byte address
          source_value = da(asm_to_char (op2));
          source_type = BYTE;
          cycles++; // source is direct ==> two clock cycles.
     }
     else if (op2[0] == '#') 
         if (is_a_nr (op2 + 1)) {       
	     source_value = asm_to_char(op2 + 1);
	     source_type = BYTE;
	 }
	 else
	     error ("Bad source constant.");
     else if (op2[0] == '@') {
          if (op2[1] == 'R' && (op2[2] == '0' ||
                                        op2[2] == '1') &&
              op2[3] == 0) {
               if (DEBUG)
                    cout << "Indirect source addr: " << (int) sfr_info[op2].address
                         << ", addr2: " << (int) da (sfr_info[op2 + 1].address)
                         << ", value = " << (int) ida (da (sfr_info[op2+ 1].address)) << endl;
               source_value = ida ( da (sfr_info[op2 +1 ].address));
               source_type = BYTE;
          }
          else
               error ("Indirect works only with R0 & R1");
     }
     else if (sfr_info.contains(op2))
          if (sfr_info[op2].type == byte) {
               source_value = da (sfr_info[op2].address);
               source_type = BYTE;
          }
          else {
               source_type = BIT;
               source_value = is_bit_set (sfr_info[op2].address);
          }
     else 
          error ("Invalid source");
     
     if (source_type + dest_type == 0)     // byte --> byte
          if (direct) {
               da (dest_addr) = source_value;
               if (DEBUG_2)
                    cout << "Direct move  ";
          }
          else 
               ida (dest_addr) = source_value;
     else if (source_type + dest_type == 2) // bit --> bit
          if (source_value)
               set_bit(dest_addr);
          else
               clr_bit (dest_addr);
     else
          error ("Wrong combination of operands for move!");
     
     zero = (source_value == 0);
     
     if (DEBUG_2) {
          cout << "Moving " << op2 << " (" << int (source_value)
               << ") to " << op1 << " (" << setbase(16)
               << (int)dest_addr << ")" << setbase(10) << "   => "
               << (int)da(dest_addr) << endl;
          cout << "dest_type = " << dest_type << ", source_type == "
               << source_type << endl;
     }
     incLN;
     cycles++;
}

void Sim8051::add (void)
{
     unsigned char dest_addr;
     char source_value;
     unsigned char negative = da (A) < 0; // & 0x80;
     if (strcmp (op1, "A")) // => not equal to 'A'
          error ("Addition needs accumulator as destination");
     else
          dest_addr = A;
     
     if (is_a_nr (op2))
          source_value = da (asm_to_char (op2));
     else if (op2[0] == 'R')
          source_value = da (sfr_info[op2].address);
     else if (op2[0] == '#')
       if (is_a_nr (op2 + 1))
          source_value = asm_to_char (op2 + 1);
       else
	 error ("Bad source constant.");
     else if (op2[0] == '@' && op2[1] == 'R' &&
              (op2[2] == '0' || op2[2] == '1') &&
              op2[3] == 0)
          source_value = ida (da (sfr_info[op2 + 1].address));
     else
          error ("invalid source");
     
//     unsigned char psw_addr = sfr_info["PSW"].address;
     if (DEBUG) {
          cerr << "Addition; source = " << int(source_value) << endl;
          cerr << "          dest_addr = " << (int)dest_addr << " = "
               << (int)da(dest_addr) << endl;
     }
     if (((da (dest_addr) & 0x0F) + (source_value & 0x0F)) & 0xF0)
          da (PSW) |= AC;
     else
          da (PSW) &= NOT_AC;
     
     da (dest_addr) += source_value;
     
     if (DEBUG)
          cerr << "result : " << (int) da (dest_addr) << endl;
     if (negative && (da (dest_addr) >= 0)) 
          da (PSW) |= CY;
     else
          da (PSW) &= NOT_CY;
     
     if ((da (dest_addr) & 0x80) && (!negative))
          da (PSW) |= OV;
     else
          da (PSW) &= NOT_OV;

     zero = da (dest_addr) == 0;
     
     if (DEBUG)
          cerr << "result : " << (int) da (dest_addr) << endl;
     if (DEBUG_2)
          cout << "Adding " << op2 << " (" << (int)source_value
               << ") to " << op1 << endl;
     incLN;
     cycles++;
}

void Sim8051::addc (void)
{
     unsigned char dest_addr;
     char source_value;
     int carry = (da (PSW) & CY ) != 0;
  
     unsigned char negative = da (A) < 0; // & 0x80;
     if (strcmp (op1, "A")) // => not equal to 'A'
          error ("Addition needs accumulator as destination");
     else
          dest_addr = A;

     if (is_a_nr (op2))
          source_value = da (asm_to_char (op2)); // Was: atoi (op2)
     else if (op2[0] == 'R')
          source_value = da (sfr_info[op2].address);
     else if (op2[0] == '#')
          source_value = asm_to_char (op2 + 1);
     else if (op2[0] == '@' && op2[1] == 'R' &&
              (op2[2] == '0' || op2[2] == '1') &&
              op2[3] == 0)
          source_value = ida (da (sfr_info[op2 + 1].address));
     else
          error ("invalid source");

     if (((da (dest_addr) & 0x0F) + ((source_value + carry) & 0x0F)) & 0xF0)
          da (PSW) |= AC;
     else
          da (PSW) &= NOT_AC;

     da (dest_addr) += source_value + carry;

     if (negative && (da (dest_addr) >= 0)) 
          da (PSW) |= CY;
     else
          da (PSW) &= NOT_CY;
 
     if ((da (dest_addr) & 0x80) && (!negative))
          da (PSW) |= OV;
     else
          da (PSW) &= NOT_OV;

     zero = da (dest_addr) == 0;
  
     if (DEBUG_2)
          cout << "Adding " << op2 << " (" << (int)source_value << ") to "
               << op1 << endl;
     incLN;
     cycles++;
}

void Sim8051::sub_borrow (void)
{

     int carry = (da (PSW) & CY) != 0;
     if (DEBUG)
          cout << "Carry = " << carry << endl;
     unsigned int Acc = 1 << 8;
     
     unsigned char dest_addr;
     if (strcmp(op1, "A"))
          error ("Substraction needs A as destination");
     dest_addr = A;
     unsigned char source = source_val();

     Acc |= (unsigned char) da (dest_addr);

     if (DEBUG)
          cout << "A = " << Acc << setbase(16) << " = " << Acc << setbase (10) << endl;

     Acc = Acc - source - carry;

     if (DEBUG)
          cout << "A = " << Acc << setbase(16) << " = " << Acc << setbase(10) << endl;
     
     if (!(Acc & (1 << 8)))
          da (PSW) |= CY;
     else
          da (PSW) &= NOT_CY;
     if ((char) Acc > da (dest_addr))
          da (PSW) |= OV;
     else
          da (PSW) &= NOT_OV;
     
     da (dest_addr) = (char) Acc;
          
     zero = da (dest_addr) == 0;
     
     incLN; 
     cycles++;
}

void Sim8051::push_on_stack (char c)
{
     da (SP)++;
     ida ( da (SP) ) = c;
}

char& Sim8051::pop_from_stack (void)
{
     return ida ( da (SP)-- );
}

void Sim8051::call (void)
{
     if (!labels.contains(op1))
          error ("Unknown call label");
     else {
          convert tmp;
          tmp.val = text->getCurLineNum () + 1;
          push_on_stack (tmp.split.low);
          push_on_stack (tmp.split.high);
          text->setCurLine (labels[op1]);
          level++;
     }
     cycles += 2;
}

void Sim8051::ret (void)
{
     convert tmp;
     tmp.split.high = pop_from_stack();
     tmp.split.low = pop_from_stack();
     text->setCurLine (tmp.val);
     level--;
     cycles += 2;
}   

void Sim8051::reti (void)
{
     convert tmp;
     da (PSW) = pop_from_stack();
     tmp.split.high = pop_from_stack();
     tmp.split.low = pop_from_stack();
     interrupt_stack.pop();
     text->setCurLine (tmp.val);
     cycles += 2;
}

void Sim8051::jump_on_zero (void)
{
     if (labels.contains (op1))
          if (zero)
               text->setCurLine (labels[op1]);
          else
               incLN;
     else
          error ("Label unknown");
     cycles += 2;
}

void Sim8051::jump_not_zero (void)
{
     if (labels.contains (op1))
          if (!zero)
               text->setCurLine (labels[op1]);
          else
               incLN;
     else
          error ("Label unknown");
     cycles += 2;
}

void Sim8051::djnz (void)
{
     unsigned char dest_addr;
     if (is_a_nr (op1))
          dest_addr = asm_to_char (op1);
     else if (sfr_info.contains(op1))
          dest_addr = sfr_info[op1].address;
     else
          error ("DJNZ: operator 1 is not valid");

     if (labels.contains(op2) || !strcmp (op2, "$")) {
          da (dest_addr)--;
          zero = da (dest_addr) == 0;
          if (!zero)
               if (strcmp (op2, "$"))
                    text->setCurLine (labels[op2]);
               else;
          else
               incLN;
     } else
          error ("Unknown label");
     cycles += 2;
}

void Sim8051::inc (void)
{
     unsigned char dest_addr;
     int direct = 1;
  
     if (is_a_nr(op1))
          dest_addr = asm_to_char (op1);
     else if (op1[0] == '@' &&
              op1[1] == 'R' &&
              (op1[2] == '0' || op1[2] == '1') &&
              op1[3] == 0)
          dest_addr = da (sfr_info[op1 + 1].address), direct = 0;
     else if (sfr_info.contains(op1) && sfr_info[op1].type == byte)
          dest_addr = sfr_info[op1].address;
     else if (!strcmp (op1, "DPTR")) {
          set_dptr (get_dptr () + 1);
          incLN;
          return;
     }
     else
          error ("Unkown operand.");

     if (direct)
          da (dest_addr)++;
     else
          ida (dest_addr)++;
     incLN;
     cycles++;
}   

void Sim8051::dec (void)
{
     unsigned char dest_addr;
     int direct = 1;
  
     if (is_a_nr(op1))
          dest_addr = asm_to_char (op1);
     else if (op1[0] == '@' &&
              op1[1] == 'R' &&
              (op1[2] == '0' || op1[2] == '1'))
          dest_addr = da (sfr_info[op1 + 1].address), direct = 0;
     else if (sfr_info.contains(op1))
          dest_addr = sfr_info[op1].address;
     else
          error ("Unkown operand.");

     if (direct)
          da (dest_addr)--;
     else
          ida (dest_addr)--;
     incLN;
     cycles++;
}

void Sim8051::digital_adjust (void)
{
     if (strcmp (op1, "A"))
          error ("operand of decimal adjust must be accumulator (A)");
     else {
          unsigned char dest_addr = A;
          if ( da (PSW) & CY ||
               ((da (dest_addr) >> 4) & 0x0F) > 9) {
               if ( (((da (dest_addr) & 0xF0) >> 4) + 0x06) & 0xF0)
                    da (PSW) |= CY;
               else
                    da (PSW) &= ~CY;
               da (dest_addr) += 0x60;
          }
          if ( da (PSW) & AC || (da (dest_addr) & 0x0F) > 9)
               da (dest_addr) += 0x06;
     }
     incLN;
     cycles++;
}

void Sim8051::setb (void)
{
     if (sfr_info.contains (op1))
          if (sfr_info[op1].type == bit)
               set_bit (sfr_info[op1].address);
          else
               error ("Not a bit address");
     else if (is_a_nr(op1))
          set_bit (asm_to_char(op1));
     else
          error ("Only symbolic names and addresses  are supported for now!");
     incLN;
     cycles++;
}

void Sim8051::clrb (void)
{
     if (sfr_info.contains (op1))
          if (sfr_info[op1].type == bit)
               clr_bit (sfr_info[op1].address);
          else if (!strcmp(op1, "A"))
               da (A) = 0x00;
          else
               error ("Not a bit address");
     else if (is_a_nr(op1))
          clr_bit (asm_to_char(op1));
     else
          error ("Only symbolic names and addresses  are supported for now!");
     incLN;
     cycles++;
}

void Sim8051::anl (void)
{
     unsigned char dest_addr;
     char source_value;
     int source_is_bit = 0, dest_is_bit = 0;
     if (is_a_nr (op1))
          dest_addr = asm_to_char(op1);
     else if (!strcmp(op1, "A"))
          dest_addr = A;
     else if (sfr_info.contains (op1) && sfr_info[op1].type == byte)
          dest_addr = sfr_info[op1].address;
     else if (sfr_info.contains(op1) && sfr_info.get(op1).type == bit)
     {
          dest_addr = sfr_info[op1].address;
          dest_is_bit = 1;
     }
     else
          error ("Not a good destination address for ANL");
  
     if (is_a_nr(op2))
          source_value = da(asm_to_char(op2));
     else if (op2[0] == '#')
          source_value = asm_to_char (op2 + 1);
     else if (op2[0] == '@') {
          if (op2[1] == 'R' && (op2[2] == '0' ||
                                        op2[2] == '1') &&
              op2[3] == 0) {
               source_value = ida ( da (sfr_info[op2 +1 ].address));
          }
          else
               error ("Indirect works only with R0 & R1");
     }
     else if (sfr_info.contains(op2))
          if (sfr_info[op2].type == bit) {
               source_is_bit = 1;
               source_value = is_bit_set (sfr_info[op2].address);
               if (DEBUG)
                    cerr << "source-value == " << (int)source_value
                         << ", address = " << setbase(16) << (int)sfr_info[op2].address << setbase(10) << endl;
          }
          else
               source_value = da (sfr_info[op2].address);
     else if (op2[0] == '/' && sfr_info.contains (op2 + 1) &&
              sfr_info[op2 + 1].type == bit) {
          source_is_bit = 1;
          source_value = !is_bit_set (sfr_info[op2+1].address);
     }
     else
          error ("Invalid source");

     if (DEBUG)
          cerr << "source is bit: " << (source_is_bit ? "true" : "false")
               << ", val = " << (int)source_value << "\n dest is bit : "
               << (dest_is_bit ? "true" : "false") << " dest_address = "
               << (int)dest_addr << " val = " << is_bit_set (dest_addr) << endl;
  
     if (source_is_bit && dest_is_bit)
          if (is_bit_set(dest_addr) && source_value)
               set_bit (dest_addr);
          else
               clr_bit (dest_addr);
     else if (!source_is_bit && !dest_is_bit)
          da (dest_addr) &= source_value;
     else
          error ("Both operands should be of the same type!");
     incLN;
     cycles++;
}

void Sim8051::orl (void)
{
     unsigned char dest_addr;
     char source_value;
     int source_is_bit = 0, dest_is_bit = 0;
     if (is_a_nr (op1))
          dest_addr = asm_to_char(op1);
     else if (!strcmp(op1, "A"))
          dest_addr = A;
     else if (sfr_info.contains (op1) && sfr_info[op1].type == byte)
          dest_addr = sfr_info[op1].address;
     else if (sfr_info.contains(op1) && sfr_info.get(op1).type == bit)
     {
          dest_addr = sfr_info[op1].address;
          dest_is_bit = 1;
     }
     else
          error ("Not a good destination address for ANL");
  
     if (is_a_nr(op2))
          source_value = da(asm_to_char(op2));
     else if (op2[0] == '#')
          source_value = asm_to_char (op2 + 1);
     else if (op2[0] == '@') {
          if (op2[1] == 'R' && (op2[2] == '0' ||
                                        op2[2] == '1') &&
              op2[3] == 0) {
               source_value = ida ( da (sfr_info[op2 +1 ].address));
          }
          else
               error ("Indirect works only with R0 & R1");
     }
     else if (sfr_info.contains(op2))
          if (sfr_info[op2].type == bit) {
               source_is_bit = 1;
               source_value = is_bit_set (sfr_info[op2].address);
               if (DEBUG)
                    cerr << "source-value == " << (int)source_value
                         << ", address = " << setbase(16) << (int)sfr_info[op2].address << setbase(10) << endl;
          }
          else
               source_value = da (sfr_info[op2].address);
     else if (op2[0] == '/' && sfr_info.contains (op2 + 1) &&
              sfr_info[op2 + 1].type == bit) {
          source_is_bit = 1;
          source_value = !is_bit_set (sfr_info[op2+1].address);
     }
     else
          error ("Invalid source");

     if (DEBUG)
          cerr << "source is bit: " << (source_is_bit ? "true" : "false")
               << ", val = " << (int)source_value << "\n dest is bit : "
               << (dest_is_bit ? "true" : "false") << " dest_address = "
               << (int)dest_addr << " val = " << is_bit_set (dest_addr) << endl;
  
     if (source_is_bit && dest_is_bit)
          if (is_bit_set(dest_addr) || source_value)
               set_bit (dest_addr);
          else
               clr_bit (dest_addr);
     else if (!source_is_bit && !dest_is_bit)
          da (dest_addr) |= source_value;
     else
          error ("Both operands should be of the same type!");
     incLN;
     cycles++;
}

void Sim8051::xorl (void)
{
     unsigned char dest_addr;
     char source_value;
     int source_is_bit = 0, dest_is_bit = 0;
     if (is_a_nr (op1))
          dest_addr = asm_to_char(op1);
     else if (!strcmp(op1, "A"))
          dest_addr = A;
     else if (sfr_info.contains (op1) && sfr_info[op1].type == byte)
          dest_addr = sfr_info[op1].address;
     else if (sfr_info.contains(op1) && sfr_info.get(op1).type == bit)
     {
          dest_addr = sfr_info[op1].address;
          dest_is_bit = 1;
     }
     else
          error ("Not a good destination address for XRL");
  
     if (is_a_nr(op2))
          source_value = da(asm_to_char(op2));
     else if (op2[0] == '#')
          source_value = asm_to_char (op2 + 1);
     else if (op2[0] == '@') {
          if (op2[1] == 'R' && (op2[2] == '0' ||
                                        op2[2] == '1') &&
              op2[3] == 0) {
               source_value = ida ( da (sfr_info[op2 +1 ].address));
          }
          else
               error ("Indirect works only with R0 & R1");
     }
     else if (sfr_info.contains(op2))
          if (sfr_info[op2].type == bit) {
               source_is_bit = 1;
               source_value = is_bit_set (sfr_info[op2].address);
               if (DEBUG)
                    cerr << "source-value == " << (int)source_value
                         << ", address = " << setbase(16) << (int)sfr_info[op2].address << setbase(10) << endl;
          }
          else
               source_value = da (sfr_info[op2].address);
     else if (op2[0] == '/' && sfr_info.contains (op2 + 1) &&
              sfr_info[op2 + 1].type == bit) {
          source_is_bit = 1;
          source_value = !is_bit_set (sfr_info[op2+1].address);
     }
     else
          error ("Invalid source");

     if (DEBUG)
          cerr << "source is bit: " << (source_is_bit ? "true" : "false")
               << ", val = " << (int)source_value << "\n dest is bit : "
               << (dest_is_bit ? "true" : "false") << " dest_address = "
               << (int)dest_addr << " val = " << is_bit_set (dest_addr) << endl;
  
     if (source_is_bit && dest_is_bit)
          if (is_bit_set(dest_addr) ^ source_value)
               set_bit (dest_addr);
          else
               clr_bit (dest_addr);
     else if (!source_is_bit && !dest_is_bit)
          da (dest_addr) ^= source_value;
     else
          error ("Both operands should be of the same type!");
     incLN;
}

void Sim8051::cpl (void)
{
     if (sfr_info.contains(op1))
          if (sfr_info[op1].type == bit)
               cpl_bit (sfr_info[op1].address);
          else if (!strcmp(op1, "A"))
               da (A) = ~da (A);
          else
               error("Wrong operand of complement");
     incLN;
     cycles++;
}
	     
void Sim8051::push (void)
{
     if (sfr_info.contains(op1) && sfr_info[op1].type == byte)
          push_on_stack (da (sfr_info[op1].address));
     else if (is_a_nr (op1))
          push_on_stack ( asm_to_char (op1));
     else
          error ("not a valid operand for push instruction");
     incLN;
     cycles += 2;
}

void Sim8051::pop (void)
{
     if (sfr_info.contains(op1) && sfr_info[op1].type == byte)
          da (sfr_info[op1].address) = pop_from_stack ();
     else if (is_a_nr (op1))
          da (asm_to_char (op1)) = pop_from_stack();
     else
          error ("not a valid operand for push instruction");
     incLN;
     cycles += 2;
}    

void Sim8051::multiply (void)
{
     if (strcmp(op1,"A") || strcmp (op2, "B"))
          error("Only possible use of mul: MUL A B");
     else {
          int tmp = da (A) * da (sfr_info["B"].address);
          da (A) = (char) (tmp & 0x00FF);
          da (sfr_info["B"].address) = (char) ((tmp & 0x0000FF00) >> 8);
          if (da(sfr_info["B"].address) != 0x00)
               da (PSW) |= OV;
          else
               da (PSW) &= ~OV;    
          clr_bit(sfr_info["C"].address);    
     }
     incLN;
     cycles += 4;
}

void Sim8051::divide (void)
{
     char res, mod;
     if (strcmp (op1, "A") || strcmp (op2, "B")) 
          error ("Only A and B can be operands of DIV");
     else {
          if (da (sfr_info["B"].address) == 0)
               da (PSW) |= OV;
          else {
               res = da (A) / da (sfr_info["B"].address);
               mod = da (A) % da (sfr_info["B"].address);
               da (A) = res;
               da (sfr_info["B"].address) = mod;
               da (PSW
                    ) &= ~OV;
          }
     }
     clr_bit (sfr_info["C"].address);
     incLN;
     cycles += 4;
}

void Sim8051::jump_on_carry ()
{
     if (is_bit_set (sfr_info["C"].address))
          if (labels.contains(op1))
               text->setCurLine (labels[op1]);
          else
               error("unknown label!");
     else
          incLN;
     cycles += 2;
}

void Sim8051::jump_no_carry ()
{
     if (!is_bit_set (sfr_info["C"].address))
          if (labels.contains(op1))
               text->setCurLine (labels[op1]);
          else
               error("unknown label!");
     else
          incLN;
     cycles += 2;
}

void Sim8051::jump_on_bit (void)
{
     unsigned char dest_addr;
     unsigned short rel;
     int ok = 1;
     if (is_a_nr (op1))
          dest_addr = asm_to_char (op1);
     else if (sfr_info.contains(op1) && sfr_info[op1].type == bit)
          dest_addr = sfr_info[op1].address;
     else
          error ("wrong operand #1"), ok = 0;

     if (is_a_nr (op2))
          rel = asm_to_char(op2);
     else if (labels.contains(op2))
          rel = labels[op2];
     else
          error ("unknown jump address!"), ok = 0;

     if (ok) 
          if (is_bit_set(dest_addr))
               text->setCurLine (rel);
          else
               incLN;
     else
          incLN;
     cycles+=2;
}   

void Sim8051::jump_no_bit (void)
{
     unsigned char dest_addr;
     unsigned short rel;
     int ok = 1;
     if (is_a_nr (op1))
          dest_addr = asm_to_char (op1);
     else if (sfr_info.contains(op1) && sfr_info[op1].type == bit)
          dest_addr = sfr_info[op1].address;
     else
          error ("wrong operand #1"), ok = 0;

     if (is_a_nr (op2))
          rel = asm_to_char(op2);
     else if (labels.contains(op2))
          rel = labels[op2];
     else
          error ("unknown jump address!"), ok = 0;

     if (ok) 
          if (!is_bit_set(dest_addr))
               text->setCurLine (rel);
          else
               incLN;
     else
          incLN;
     cycles += 2;
}   

void Sim8051::jump_on_bit_and_clear (void)
{
     unsigned char dest_addr;
     unsigned short rel;
     int ok = 1;
     if (is_a_nr (op1))
          dest_addr = asm_to_char (op1);
     else if (sfr_info.contains(op1) && sfr_info[op1].type == bit)
          dest_addr = sfr_info[op1].address;
     else
          error ("wrong operand #1"), ok = 0;

     if (is_a_nr (op2))
          rel = asm_to_char(op2);
     else if (labels.contains(op2))
          rel = labels[op2];
     else
          error ("unknown jump address!"), ok = 0;

     if (ok) 
          if (is_bit_set(dest_addr)) {
               clr_bit(dest_addr);
               text->setCurLine (rel);
          }
          else
               incLN;
     else
          incLN;
     cycles += 2;
}   

void Sim8051::nop (void) {
     incLN;
     cycles++;
}

void Sim8051::zilch (void) {
     incLN;
}

void Sim8051::rl (void)
{
     if (strcmp(op1, "A"))
          error ("Only A can be rotated!");
     else {
          int tmp = da (A);
          int last_bit = (da (A) & 0x80) != 0;
          tmp <<= 1;
          if (last_bit)
               tmp |= 0x01;
          else
               tmp &= 0xFE;
          da (A) = tmp;
     }
     incLN;
     cycles++;
}

void Sim8051::rlc (void)
{
     if (strcmp(op1, "A"))
          error ("Only A can be rotated!");
     else {
          int tmp = da (A);
          int last_bit = is_bit_set(sfr_info["C"].address);
          if ((da (A) & 0x80) != 0)
               set_bit(sfr_info["C"].address);
          else
               clr_bit(sfr_info["C"].address);
          tmp <<= 1;
          if (last_bit)
               tmp |= 0x01;
          else
               tmp &= 0xFE;
          da (A) = tmp;
     }
     incLN;
     cycles++;
}

void Sim8051::rr (void)
{
     if (strcmp(op1, "A"))
          error ("Only A can be rotated!");
     else {
          int tmp = da (A);
          int last_bit = (da (A) & 0x01) != 0;
          tmp >>= 1;
          if (last_bit)
               tmp |= 0x80;
          else
               tmp &= 0x7F;
          da (A) = tmp;
     }
     incLN;
     cycles++;
}  

void Sim8051::rrc (void)
{
     if (strcmp(op1, "A"))
          error ("Only A can be rotated!");
     else {
          int tmp = da (A);
          int last_bit = is_bit_set(sfr_info["C"].address);
          if ((da (A) & 0x01) != 0)
               set_bit (sfr_info["C"].address);
          else
               clr_bit (sfr_info["C"].address);
          tmp >>= 1;
          if (last_bit)
               tmp |= 0x80;
          else
               tmp &= 0x7F;
          da (A) = tmp;
     }
     incLN;
     cycles++;
}  

void Sim8051::swap (void)
{
     if (strcmp(op1,"A"))
          error ("Only Accumulator can be swapped");
     else {
          unsigned char high, low;
          high = da (A);
          low = (high & 0x0F) << 4;
          high = ((high & 0xF0) >> 4) & 0x0f; // to avoid problems
          da (A) = high | low;
     }
     incLN;
     cycles++;
}

void Sim8051::exchange (void)
{
     if (strcmp("A",op1))
          error ("XCH can only have the Acc as destination");
     else {
          char tmp = da (A);
          char* source_address;
          if (is_a_nr(op2))
               source_address = & da (asm_to_char(op2));
          else if (op2[0] == '#')
               error ("You can't XCH A with data!");
          else if (op2[0] == '@') {
               if (op2[1] == 'R' && (op2[2] == '0' ||
                                             op2[2] == '1') &&
                   op2[3] == 0) {
                    source_address = & ida ( da (sfr_info[op2 +1 ].address));
               }
               else
                    error ("Indirect works only with R0 & R1");
          }
          else if (sfr_info.contains(op2)) 
               source_address = & da (sfr_info[op2].address);
          else
               error ("Invalid source");
          da (A) = *source_address;
          *source_address = tmp;
     }
     incLN;
     cycles++;
}

void Sim8051::exchange_nibble (void)
{
     if (strcmp("A", op1))
          error ("XCHD only works with Acc as destination.");
     else {
          char tmp = da (A);
          char* source_address;
          if (op2[0] == '@') {
               if (op2[1] == 'R' && (op2[2] == '0' ||
                                             op2[2] == '1') &&
                   op2[3] == 0) {
                    source_address = & ida ( da (sfr_info[op2 +1 ].address));
               }
               else
                    error ("Indirect works only with R0 & R1");
          }
          else
               error ("Invalid source");
          da (A) &= 0x0F;
          da (A) |= (*source_address & 0xF0);
          *source_address &= 0x0f;
          *source_address |= tmp & 0xF0;
     }
     incLN;
     cycles++;
}	  

void Sim8051::cmp_jump_on_not_equal (void)
{
     char dest, source;
     if (!strcmp("A", op1) ||
         !strcmp("Acc", op1))
          dest = da(A);
     else if (op1[0] == 'R' &&
              (op1[1] >= '0' && op1[1] <= '7') &&
              op1[2] == 0)
          dest = da (sfr_info[op1].address);
     else if (op1[0] == '@' &&
              op1[1] == 'R' &&
              (op1[2] == '1' || op1[2] == '0') &&
              op1[3] == 0)
          dest = ida ( da (sfr_info[op1 + 1].address) );
     else
          error ("CJNE Wrong destination parameter");

     if (is_a_nr (op2))
          source = da (asm_to_char(op2));
     else if (op2[0] == '#' && is_a_nr (op2 + 1))
          source = asm_to_char (op2 + 1);
     else
          error ("CJNE Bad source parameter");

     if (source != dest)
          if (labels.contains(op3))
               text->setCurLine (labels[op3]);
          else
               error ("Unknown label!");
     else
          incLN;

     if (DEBUG)
          cout << "CJNE called. dest = " << (int)dest << ", source = "
               << (int)source << ", label = " << op3
               << " of which line nr = " << labels.get (op3) << endl;
     cycles+=2;
}

void Sim8051::createMT (void)
{
     if (DEBUG)
          cerr << "Creating mnemonic table... ";
     mnemonic["ADD"] = &Sim8051::add; 
     mnemonic["ADDC"] = &Sim8051::addc;
     mnemonic["SUBB"] = &Sim8051::sub_borrow;
     mnemonic["INC"] = &Sim8051::inc;  
     mnemonic["DEC"] = &Sim8051::dec;
     mnemonic["MUL"] = &Sim8051::multiply;  
     mnemonic["DIV"] = &Sim8051::divide;
     mnemonic["DA"] = &Sim8051::digital_adjust;  
     mnemonic["ANL"] = &Sim8051::anl;
     mnemonic["ORL"] = &Sim8051::orl;  
     mnemonic["XRL"] = &Sim8051::xorl;
     mnemonic["CLR"] = &Sim8051::clrb;  
     mnemonic["CPL"] = &Sim8051::cpl;
     mnemonic["RL"] = &Sim8051::rl;  
     mnemonic["RLC"] = &Sim8051::rlc;
     mnemonic["RR"] = &Sim8051::rr;  
     mnemonic["RRC"] = &Sim8051::rrc;
     mnemonic["SWAP"] = &Sim8051::swap;  
     mnemonic["MOV"] = &Sim8051::move;
     mnemonic["MOVX"] = &Sim8051::unknown;  
     mnemonic["PUSH"] = &Sim8051::push;
     mnemonic["POP"] = &Sim8051::pop;  
     mnemonic["XCH"] = &Sim8051::exchange;
     mnemonic["XCHD"] = &Sim8051::exchange_nibble;  
     mnemonic["SETB"] = &Sim8051::setb;  
     mnemonic["ACALL"] = &Sim8051::call;
     mnemonic["LCALL"] = &Sim8051::call;
     mnemonic["CALL"] = &Sim8051::call;
     mnemonic["RET"] = &Sim8051::ret;
     mnemonic["RETI"] = &Sim8051::reti;  
     mnemonic["AJMP"] = &Sim8051::jump;          
     mnemonic["LJMP"] = &Sim8051::jump;          
     mnemonic["SJMP"] = &Sim8051::jump;
     mnemonic["JMP"] = &Sim8051::jump;           
     mnemonic["JZ"] = &Sim8051::jump_on_zero;
     mnemonic["JNZ"] = &Sim8051::jump_not_zero;  
     mnemonic["JC"] = &Sim8051::jump_on_carry;
     mnemonic["JNC"] = &Sim8051::jump_no_carry;
     mnemonic["JB"] = &Sim8051::jump_on_bit;
     mnemonic["JNB"] = &Sim8051::jump_no_bit;  
     mnemonic["JBC"] = &Sim8051::jump_on_bit_and_clear;
     mnemonic["CJNE"] = &Sim8051::cmp_jump_on_not_equal;  
     mnemonic["DJNZ"] = &Sim8051::djnz;
     mnemonic["NOP"] = &Sim8051::nop;
     mnemonic["END"] = &Sim8051::end;
     mnemonic[""] = &Sim8051::zilch; // for labels on an empty line
     mnemonic["ORG"] = &Sim8051::zilch;

     mnemonic["MOVC"] = &Sim8051::no_movc;
  
     if (DEBUG)
          cerr << "Done.\n";
}

void Sim8051::int0_int (void)
{
//     cerr << "int 0 interrupt generated " << (da (IE) & 0x80) << ";  " << (da (IE) & 0x01) << endl;
//     interrupt_stack.show();
     if ((da(IE) & 0x80) && (da (IE) & 0x01)) {
          if (!interrupt_stack.contains_higher (external_0,
                                                (da(IP) & 0x01) != 0)) {
               if (is_bit_set (sfr_info["IT0"].address)) // neg. flank
                    clr_bit (sfr_info["IE0"].address);
               if (!labels.contains ("0003H"))
                    error ("ISR for external 0 should be indicated by label (!) 0003H");
               else {
                    convert tmp;
                    tmp.val = text->getCurLineNum();
                    push_on_stack (tmp.split.low);
                    push_on_stack (tmp.split.high);
                    push_on_stack (da (PSW));
                    interrupt_stack.push (external_0,
                                          is_bit_set (sfr_info["IP.0"].address));
                    text->setCurLine (labels["0003H"]);
               }
          }
     }
}

void Sim8051::int1_int (void)
{
     if ((da(IE) & 0x80) && ((da (IE) & 0x04))) {
          if (!interrupt_stack.contains_higher (external_1,
                                                (da(IP) & 0x04) != 0)) {
               if (da (TCON) & 0x04) // neg. flank
                    clr_bit (sfr_info["IE1"].address);
               if (!labels.contains ("0013H"))
                    error ("ISR for external 0 should be indicated by label (!) 0013H");
               else {
                    convert tmp;
                    tmp.val = text->getCurLineNum();
                    push_on_stack (tmp.split.low);
                    push_on_stack (tmp.split.high);
                    push_on_stack (da (PSW));
                    interrupt_stack.push (external_1,
                                          is_bit_set (sfr_info["IP.2"].address));
                    text->setCurLine (labels["0013H"]);
               }
          }
     }
}

void Sim8051::tc_0_int (void)
{
     if (is_bit_set (sfr_info["EA"].address) &&
         is_bit_set (sfr_info["ET0"].address) &&
         !interrupt_stack.contains_higher (timer_0, is_bit_set (sfr_info["IP.1"].address))) {
          clr_bit (sfr_info["TF0"].address);
          if (!labels.contains("000BH"))
               error("ISR should be indicated by label (!) 000BH:");
          else {
               convert tmp;
               tmp.val = text->getCurLineNum();
               push_on_stack (tmp.split.low);
               push_on_stack (tmp.split.high);
               push_on_stack (da (PSW));
               interrupt_stack.push (timer_0, is_bit_set (sfr_info["IP.1"].address));
               text->setCurLine (labels["000BH"]);
          }
     }
}
     
void Sim8051::tc_1_int (void)
{
     if (is_bit_set (sfr_info["EA"].address) &&
         is_bit_set (sfr_info["ET1"].address) &&
         !interrupt_stack.contains_higher (timer_1, is_bit_set (sfr_info["IP.3"].address))) {
          clr_bit(sfr_info["TF1"].address);
          if (!labels.contains("001BH"))
               error("ISR should be indicated by label (!) 001BH:");
          else {
               convert tmp;
               tmp.val = text->getCurLineNum ();
               push_on_stack (tmp.split.low);
               push_on_stack (tmp.split.high);
               push_on_stack (da (PSW));
               interrupt_stack.push (timer_1, is_bit_set (sfr_info["IP.3"].address));
               text->setCurLine (labels["001BH"]);
          }
     }
}

void Sim8051::inc_tc_0 (void)
{
     int mode = da (sfr_info["TMOD"].address) & 0x03;
     unsigned char tl = sfr_info["TL0"].address, th = sfr_info["TH0"].address;
     unsigned short total;
     switch (mode) {
     case 0:
          da (tl)++;
          if (da (tl) == 32) {
               da(tl) = 0;
               da (th)++;
               if (da(th) == 0)
                    set_bit (sfr_info["TF0"].address);
          }
          break;
     case 1:
          total = ((unsigned char)da (th) << 8) | (unsigned char)da(tl);
          total++;
          if (total == 0)
               set_bit(sfr_info["TF0"].address);
          da (th) = (char) ((total >> 8) & 0x00FF);
          da (tl) = (char) (total & 0x0FF);
          break;
     case 2:
          da (tl) ++;
          if (da (tl) == 0) {
               da (tl) = da (th);
               set_bit(sfr_info["TF0"].address);
          }
          break;
     }
}    

void Sim8051::inc_tc_1 (void)
{
     int mode = (da (sfr_info["TMOD"].address) >> 4) & 0x03;
     unsigned char tl = sfr_info["TL1"].address, th = sfr_info["TH1"].address;
     unsigned short total;
     switch (mode) {
     case 0:
          da (tl)++;
          if (da (tl) == 32) {
               da(tl) = 0;
               da (th)++;
               if (da(th) == 0)
                    set_bit (sfr_info["TF1"].address);
          }
          break;
     case 1:
          total = ((unsigned char)da (th) << 8) | (unsigned char)da(tl);
          total++;
          if (total == 0)
               set_bit(sfr_info["TF1"].address);
          da (th) = (char) ((total >> 8) & 0x00FF);
          da (tl) = (char) (total & 0x0FF);
          break;
     case 2:
          da (tl) ++;
          if (da (tl) == 0) {
               da (tl) = da (th);
               set_bit(sfr_info["TF1"].address);
          }
          break;
     }
}

void Sim8051::timer_counter(void)
{
     static int prev_t0, prev_t1;
     
     // timer 1 first
     
     if (da (TMOD) & 0x80) // GATE is set
          if (is_bit_set (sfr_info["P3.3"].address) && (da (TCON) & 0x40)) // t0 runs
               inc_tc_1();
          else;
     else if (is_bit_set (sfr_info["TR1"].address))
          if (da (TMOD) & 0x40 && ( da(TCON) & 0x40)) // counter && tr1
               if (is_bit_set (sfr_info["P3.5"].address) && !(prev_t1)) { // pos edge
                    inc_tc_1();
                    prev_t1 = is_bit_set (sfr_info["P3.5"].address);
               }
               else
                    prev_t1 = is_bit_set (sfr_info["P3.5"].address);
          else if (is_bit_set (sfr_info["TR1"].address)) // timer : count every cycle
               inc_tc_1();

     // Then timer 0

     if (da (TMOD) & 0x08) // GATE is set
          if (is_bit_set (sfr_info["P3.2"].address) && (da (TCON) & 0x10)) // t0 runs
               inc_tc_0();
          else;
     else if (is_bit_set (sfr_info["TR0"].address))
          if (da (TMOD) & 0x04 && ( da(TCON) & 0x10)) // counter && tr0
               if (is_bit_set (sfr_info["P3.4"].address) && !(prev_t0)) { // pos edge
                    inc_tc_0();
                    prev_t0 = is_bit_set (sfr_info["P3.4"].address);
               }
               else
                    prev_t0 = is_bit_set (sfr_info["P3.4"].address);
          else if (is_bit_set (sfr_info["TR0"].address)) // timer : count every cycle
               inc_tc_0();
  
}

void Sim8051::external_interrupts (void)
{
     // check int 0 first:

     if (da (TCON) & 0x01) // IT0 is set; flank
          if (old_EXT0 && ((da (P3) & 0x04) == 0))  // neg flank
               da (TCON) |= 0x02; // set IE0
          else;
     else {
          if ((da (P3) & 0x04) == 0) // EXT 0 is low
               da (TCON) |= 0x02; // set IE0
          else // EXT 0 is high
               da (TCON) &= 0xFD; // clear IE0
     }
     old_EXT0 = (da (P3) & 0x04);

     // then check int 1:

     if (da (TCON) & 0x04) // IT1 is set
          if (old_EXT1 && ((da (P3) & 0x08) == 0))
               da (TCON) |= 0x08; // set IE1
          else;
     else {
          if ((da (P3) & 0x08) == 0) // EXT 1 is low
               da (TCON) |= 0x08; // set IE1
          else
               da (TCON) &= 0xF7; // clear IE1
     }
     old_EXT1 = (da (P3) & 0x08);
}

void Sim8051::reset (void)
{
     if (da (PCON) & 0x01) 
          da (PCON) &= 0xFE;
     else {
          text->setCurLine (0);
          da (sfr_info["A"].address) = 0x00;
          da (sfr_info["B"].address) = 0x00;
          da (sfr_info["PSW"].address) = 0x00;
          da (sfr_info["SP"].address) = 0x07;
          da (sfr_info["DPL"].address) = 0x00;
          da (sfr_info["DPH"].address) = 0x00;
          da (sfr_info["P0"].address) = 0xFF;
          da (sfr_info["P1"].address) = 0xFF;
          da (sfr_info["P2"].address) = 0xFF;
          da (sfr_info["P3"].address) = 0xFF;
          da (sfr_info["IP"].address) = 0x00;
          da (sfr_info["IE"].address) = 0x00;
          da (sfr_info["TMOD"].address) = 0x00;
          da (sfr_info["TCON"].address) = 0x00;
          da (sfr_info["TLO"].address) = 0x00;
          da (sfr_info["TH0"].address) = 0x00;
          da (sfr_info["TL1"].address) = 0x00;
          da (sfr_info["TH1"].address) = 0x00;
          da (sfr_info["SCON"].address) = 0x00;
          da (sfr_info["PCON"].address) = 0x00;
          sfr_info.free();
          create_sfr_table ();
          old_EXT0 = 1;
          interrupt_stack.free();
          text->createLT(&labels);
          cycles = 0;
          noOutput = 0;
          results->update();
          info->setVal (cycles);
          err->clear ();
          level = 0;
          stopLevel = 0;
     }
}

void Sim8051::set_parity(void)
{
     unsigned short a = da (A);
     unsigned ones = 0;
     for (int i = 0; i < 8; i++) {
          if (a & 0x01)
               ones++;
          a >>= 1;
     }
     if (ones & 0x01) // odd nr of set bits
          da (PSW) |= 0x01;
     else
          da (PSW) &= 0xFE;
}

void Sim8051::check_system (void) // look for interrupts, changes in mem banks, etc.
{
     if ((old_PSW && 0x14) != (da (PSW) && 0x14)) {
         create_r_registers(); // if necessary
         old_PSW = da (PSW);
         if (!noOutput)
              results->update();
     }
     if (!(da (PCON) & 0x02)) { // don't check those if power down.
          timer_counter();
          external_interrupts ();
     }
     if (da (TCON) & 0xAA) // if interrupt resume from idle
          da (PCON) &= 0xFE;
     if ((da (TCON) & 0x02) != 0)
          int0_int ();
     if (is_bit_set (sfr_info["TF0"].address))
          tc_0_int();
     if ((da (TCON) & 0x08) != 0)
          int1_int ();
     if (is_bit_set (sfr_info["TF1"].address))
          tc_1_int();
     
}
