/*
 * This file is part of the YASMIN 8051 simulator project
 *
 * Copyright (c) 2000 by Gwenn Englebienne
 *
 * This software is free software; it is protected by the GPL. You can 
 * modify and/or redistribute it under the terms specified by the GNU
 * Public Licence version 2.
 *
 * Be advised that this software comes with no warranty at all, not
 * even the implied warranty of merchantability or fitness for a 
 * particular purpose. 
 *
 * For more information, you could look at the COPYING file, or have a 
 * look at the free software foundation's web site: http://www.fsf.org
 *
 */
 
#ifndef Sim8051_H
#define Sim8051_H

#include <qgrid.h>
#include <qmultilineedit.h>
#include <qscrollbar.h>
#include <stdio.h>
#include "memory.h"
#include "debugwidget.h"
#include "general.h"
#include "int_stack.h"
#include "memview.h"
#include "memhex.h"
#include "portsview.h"

class ShowInfo : public QGrid {
public:
     ShowInfo (QWidget *parent = 0, const char *name = 0)
          : QGrid (2, parent, name) {
          lbl = new QLabel ("Clock cycles: ", this);
          val = new QLabel ("0", this);
          lbl2 = new QLabel ("Elapsed time: ", this);
          val2 = new QLabel ("0", this);
     }

     void setVal (unsigned int v) {
          char buf[11];
          snprintf (buf, 11, "%d", v);
          val->setText (buf);
          snprintf (buf, 11, "%d ns.", v * 750);
          val2->setText (buf);
     }
     
protected:
     QLabel *lbl, *val, *lbl2, *val2;
};

class Sim8051 : public QGrid {
     
     Q_OBJECT
     

 public:
     Sim8051 (QWidget *parent = 0, const char *name = 0);
     ~Sim8051 ();
     

 public slots:
     void step (void);
     void stepIn();
     void stepOut ();
     void stepOver ();
     void run();
     void runFast();
     void stop ();
     void reset();
     void load ();
     void save ();
     void saveAs ();
     void newFile ();
     void check ();
     void changeScroll (int y);
     
 protected:

     typedef void (Sim8051::*func) (void);
     
     DebugWidget *text;
     //     CommandBar *cmd;
     QTimer *timer;
     MemView *results;
     MemHex *overview;
     ShowInfo *info;
     PortsView *pv;
     QMultiLineEdit *err;
     QScrollBar *sb;

     char label[20], instr[20], op1[20], op2[20], op3[20];
     hashTable <unsigned int> labels;
     hashTable <func> mnemonic;
     isr_stack interrupt_stack;
     int zero, noOutput, specialStep;
     unsigned char old_PSW, old_EXT0, old_EXT1;
     unsigned int cycles, level, stopLevel;
                                    
     void fetch();
     void decode();
     void execute();
     void error (const char *s);
     char source_val (void);
     int varNameOk (char *);
     
     // And here comes the long list of mnemonics the 8051 can handle.
     // Yes, I know there are more, but I'm not done yet.

     void unknown               (void);
     void no_movc               (void);
     void jump                  (void);
     void move                  (void);
     void add                   (void);
     void addc                  (void);
     void sub_borrow            (void);
     void push_on_stack         (char c);
     char & pop_from_stack      (void);
     void call                  (void);
     void ret                   (void);
     void reti                  (void);
     void jump_on_zero          (void);
     void jump_not_zero         (void);
     void djnz                  (void);
     void inc                   (void);
     void dec                   (void);
     void digital_adjust        (void);
     void setb                  (void);
     void clrb                  (void);
     void anl                   (void);
     void orl                   (void);
     void xorl                  (void);
     void cpl                   (void);
     void push                  (void);
     void pop                   (void);
     void multiply              (void);
     void divide                (void);
     void jump_on_carry         (void);
     void jump_no_carry         (void);
     void jump_on_bit           (void);
     void jump_no_bit           (void);
     void jump_on_bit_and_clear (void);
     void nop                   (void);
     void rl                    (void);
     void rlc                   (void);
     void rr                    (void);
     void rrc                   (void);
     void swap                  (void);
     void exchange              (void);
     void exchange_nibble       (void);
     void cmp_jump_on_not_equal (void);
     void end                   (void);

// And the functions to fix the details...
     
     void zilch               (void);
     void createMT            (void); // create mnemonic table.
     void int0_int            (void);
     void int1_int            (void);
     void tc_0_int            (void);
     void tc_1_int            (void);
     void inc_tc_0            (void);
     void inc_tc_1            (void);
     void external_interrupts (void);
     void timer_counter       (void);
     void set_parity          (void);
     void check_system        (void);
};

#endif // Sim8051_H
