/*
 * This file is part of the YASMIN 8051 simulator project
 *
 * Copyright (c) 2000 by Gwenn Englebienne
 *
 * This software is free software; it is protected by the GPL. You can 
 * modify and/or redistribute it under the terms specified by the GNU
 * Public Licence version 2.
 *
 * Be advised that this software comes with no warranty at all, not
 * even the implied warranty of merchantability or fitness for a 
 * particular purpose. 
 *
 * For more information, you could look at the COPYING file, or have a 
 * look at the free software foundation's web site: http://www.fsf.org
 *
 */
 
#include "debugwidget.h"
#include <qfiledialog.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qimage.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include "general.h"
#include <qsizepolicy.h>

#define LINENR           30
#define BREAK            18
#define CURRENT          18
#define SEP              10

DebugWidget::DebugWidget(QWidget *parent, const char *name,
			 const char *fileName)
     : QWidget (parent, name)
{
     setMinimumSize (200, 100);
     setFocusPolicy (QWidget::ClickFocus);
     setFocus();
     defFont = new QFont ("lucidatypewriter", 12);
     metrics = new QFontMetrics (*defFont);
     lh = int (metrics->height() * 1.3);
     cursorOn = 0;
     changed = 0;
     showln = 1;
     showbp = 1;
     showcl = 1;
     currentLine = 0;
     indent = LINENR + BREAK + CURRENT + SEP;
     
// cursor blinking:
     blinker = new QTimer (this);
     connect (blinker, SIGNAL (timeout()),
              this, SLOT (blinkCursor()));
//     blinker->start(250);
     
// loading pixmaps and setting up painters
     setBackgroundColor (white);
     currentPix = new QImage ("smallblue.xpm");
     breakPix = new QImage ("break.xpm");
     pix = new QPixmap (width(), height());
     paint = new QPainter (pix);
     paint->setBackgroundColor (white);
     cpaint = new QPainter (this);
     cpaint->setRasterOp (Qt::XorROP);
     cpaint->setPen (Qt::white);
     cpaint->setBackgroundColor (white);

     if (fileName) {
          filename = new char[strlen(fileName) + 1];
          strcpy (filename, fileName);
          loadFile ();
     }
     else {
          txt.insertLine (0);
          filename = 0;
     }
}

DebugWidget::~DebugWidget()  {
     delete cpaint;
     delete paint;
     delete pix;
}

void DebugWidget::blinkCursor ()
{
     int x, y;
     try {
          x = metrics->width(txt.getChars (cursor.y()), cursor.x());
          y = lh * (1 + cursor.y() - origin.y());
     } catch (char * c) {
          x = 0; y = 10;
     }
     cpaint->drawLine (x + indent, y + 10, x + indent, y + 10 - lh);
     
     cursorOn =  1 - cursorOn;
}

void DebugWidget::eraseCursor ()
{
     if (cursorOn) 
	  blinkCursor();
}

void DebugWidget::draw()
{
     QColor pastelRed (255, 255, 255); //pastelRed(255, 240, 240);
     QColor pastelGreen (255, 255, 255); //pastelGreen(240, 255, 240);
     for (int i = 0; i < (height() / lh) + 2; i++)
          if ((origin.y() + i) % 2)
               paint->fillRect (0, int((i - 0.2) * lh), width(),
                                lh, pastelRed);
          else
               paint->fillRect (0, int((i - 0.2) * lh), width(),
                                lh, pastelGreen);
     paint->setFont (*defFont);	  
     paint->fillRect (0, 0, indent - SEP, height(), lightGray);
     for (int i = (origin.y() > 0 ? origin.y() - 1 : origin.y());
          i < txt.getNumLines() &&
               i < origin.y() + height() / lh; i++) {
          if (showcl && i == currentLine)
               paint->drawImage (LINENR, 10 + (i - origin.y()) * lh
                                 + (lh - currentPix->height())/2, *currentPix);
          if (showbp && txt.isBreakSet (i))
               paint->drawImage (LINENR + BREAK, 10 + (i - origin.y()) * lh
                                 + (lh - breakPix->height()) / 2, *breakPix);
          paint->drawText (2, 10 + (1 + i - origin.y()) * lh,
                           gcvt(i, 50, str));
          paint->drawText (indent, 10 + (1 + i - origin.y()) * lh,
                           txt.getChars (i));
     }
     paint->drawLine (indent - SEP, 0, indent - SEP, height());
     paint->setPen (QPen (Qt::black, 2));
     paint->drawLine (0, 0, 0, height());
     paint->drawLine (0, 0, width(), 0);
     paint->setPen (QPen (Qt::white, 2));
     paint->drawLine (0, height()-1, width()-1, height()-1);
     paint->drawLine (width()-1, 0, width()-1, height()-1);
     bitBlt (this, 0, 0, pix, 0, 0, width(), height(), Qt::CopyROP);
     paint->setPen (QPen (Qt::black));
/*     if (selStart != cursor) {
          paint->setRasterOp (Qt::XorROP);
          if (selStart.y() == cursor.y()) {
               paint->fillRect (indent + metrix.width (text->getchars (cursor.y()), selStart.x()),
                                10 + (selstart.y() - origin.y()) * lh,
*/                              
                                
                                
}

void DebugWidget::drawLine ()
{
     drawLine (cursor.y());
}

void DebugWidget::drawLine(unsigned int i)
{
     if (i < txt.getNumLines()) {
          if (i == currentLine &&
              i > origin.y() - 2 + height() / lh || i < origin.y()) {
               origin.setY(i - height() / (lh + lh));
               if (origin.y() < 0) 
                    origin.setY(0);
               emit originChange (origin.y());
               paintEvent (0);
               return;
          }
          QColor pastelRed (255, 255, 255); //pastelRed(255, 240, 240);
          QColor pastelGreen (255, 255, 255); //pastelGreen(240, 255, 240);
          if (!(i % 2))
               paint->fillRect (0, (i - origin.y() + 0.8) * lh,
                                width(), lh, pastelRed);
          else
               paint->fillRect (0, (i - origin.y() + 0.8) * lh, width(),
                                lh, pastelGreen);
          paint->fillRect (0, (i - origin.y() + 0.8) * lh,
                           indent - SEP, lh, lightGray);
          paint->setFont (*defFont);
          if (showcl && i == currentLine)
               paint->drawImage (LINENR, 10 + (i - origin.y()) * lh
                                 + (lh - currentPix->height())/2, *currentPix);
          if (showbp && txt.isBreakSet (i))
               paint->drawImage (LINENR + BREAK, 10 + (i - origin.y()) * lh
                                 + (lh -  breakPix->height()) / 2, *breakPix);
          paint->drawText (2, 10 + (1 + i - origin.y()) * lh,
                           gcvt(i, 50, str));
          paint->drawText (indent, 10 + (1 + i - origin.y()) * lh,
                           txt.getChars (i));
          paint->setPen (Qt::black);
          paint->drawLine (indent-SEP, (i - origin.y() + 0.8) * lh,
                           indent-SEP, (i - origin.y() + 1.8) * lh);
          
          paint->setPen (QPen (Qt::black, 2));
          paint->drawLine (0, 0, 0, height());
          paint->drawLine (0, 0, width(), 0);
          paint->setPen (QPen (Qt::white, 2));
          paint->drawLine (0, height()-1, width()-1, height()-1);
          paint->drawLine (width()-1, 0, width()-1, height()-1);
          paint->setPen (QPen (Qt::black));	  
          
          bitBlt (this, 0, (i - origin.y() + 0.8) * lh,
                  pix, 0, (i - origin.y() + 0.8) * lh,
                  width(), lh, Qt::CopyROP);
     }
}

void DebugWidget::paintEvent (QPaintEvent *e)
{
     draw();
}

void DebugWidget::gotoNextLine ()
{
/*     QColor pastelRed (255, 255, 255); //pastelRed(255, 240, 240);
       QColor pastelGreen (255, 255, 255); //pastelGreen(240, 255, 240); */
     cpaint->setRasterOp (Qt::CopyROP);
/*     if (!(currentLine % 2))
          cpaint->fillRect (LINENR, 10 + (currentLine - origin.y()) * lh,
                            currentPix->width(), lh, pastelRed);
                            else */
          cpaint->fillRect (LINENR, 10 + (currentLine - origin.y()) * lh,
                            currentPix->width(), lh, lightGray);
     currentLine += 1;
     cpaint->drawImage (LINENR, 10 + (currentLine - origin.y()) * lh
                        + (lh - currentPix->height())/2, *currentPix);
     cpaint->setRasterOp (Qt::XorROP);
     if (currentLine > origin.y() - 2 + height() / lh ||
         currentLine < origin.y()) {
          origin.setY(currentLine - height() / (lh + lh));
          if (origin.y() < 0)
               origin.setY(0);
          emit originChange (origin.y());
          paintEvent (0);
     }
};
     
void DebugWidget::setCurLine (unsigned int nr)
{
     cpaint->setRasterOp (Qt::CopyROP);
/*     QColor pastelRed (255, 255, 255); //pastelRed(255, 240, 240);
     QColor pastelGreen (255, 255, 255); //pastelGreen(240, 255, 240);
     if (!(currentLine % 2))
          cpaint->fillRect (LINENR, 10 + (currentLine - origin.y()) * lh,
                            currentPix->width(), lh, pastelRed);
                            else */
          cpaint->fillRect (LINENR, 10 + (currentLine - origin.y()) * lh,
                            currentPix->width(), lh, lightGray);
     currentLine = nr;
     clchanged = true;
     cpaint->drawImage (LINENR, 10 + (currentLine - origin.y()) * lh
                        + (lh - currentPix->height())/2, *currentPix);
     cpaint->setRasterOp (Qt::XorROP);
     if (currentLine > origin.y() - 2 + height() / lh ||
         currentLine < origin.y()) {
          origin.setY(currentLine - height() / (lh + lh));
          if (origin.y() < 0)
               origin.setY(0);
          emit originChange (origin.y());
          paintEvent (0);
     }
};

void DebugWidget::keyPressEvent (QKeyEvent *e)
{
     eraseCursor();
     char *tmp;
     long int res;
     
     switch (e->key()) {
     case Key_Return:
          if (cursor.y() < currentLine)
               currentLine++, clchanged = 1;
          tmp = txt.getLeadingBlanks(cursor.y());
          txt.splitLine (cursor.y(), cursor.x());
          cursor.setY(cursor.y() + 1);
          if (tmp)
               txt.insertChars (tmp, cursor.y(), 0);
          cursor.setX(strlen(tmp));
          paintEvent(0);
          changed = 1;
          delete [] tmp;
          break;
     case Key_Tab:
          txt.insertChars (" ", cursor.y(), cursor.x ());
          cursor.setX (cursor.x() + 1);
          for (int i = cursor.x(); i % 8; i++) {
               txt.insertChars (" ", cursor.y(), cursor.x());
               cursor.setX(cursor.x()+1);
          }
//          txt.insertChars ("        ", cursor.y(), cursor.x());
//          cursor.setX(cursor.x() + 8);
          drawLine();
          changed = 1;
          break;
     case Key_Backspace:
	  if (cursor.x() > 0) {
           tmp = txt.getChars (cursor.y());
           if (tmp[cursor.x() - 1] == ' ') {
                cursor.setX (cursor.x() - 1);
                txt.deleteChar (cursor.y(), cursor.x());
                while (tmp[cursor.x() - 1] == ' ' && cursor.x() % 8) {
                     cursor.setX(cursor.x() - 1);
                     txt.deleteChar (cursor.y(), cursor.x());
                }
           }
           else {
                cursor.setX (cursor.x() - 1);
                txt.deleteChar (cursor.y(), cursor.x());
           }
	       drawLine();           
	  }
	  else if (cursor.y() > 0) {
	       cursor.setY(cursor.y() - 1);
	       cursor.setX (txt.lineLength(cursor.y()));
	       txt.mergeLines (cursor.y());
	       if (cursor.y() < currentLine)
		    currentLine--, clchanged = 1;
	       draw();
	  }
      changed = 1;
	  break;
     case Key_Delete:
	  if (cursor.x() < txt.lineLength(cursor.y())) {
	       txt.deleteChar (cursor.y(), cursor.x());
	       drawLine();
	  }
      else if (cursor.y() < txt.getNumLines() - 1) {
           txt.mergeLines (cursor.y());
           if (cursor.y() < currentLine)
                currentLine--, clchanged = 1;
           draw();
      }
      changed = 1;
	  break;
     case Key_Up:
	  if (cursor.y() > 0)
	       cursor.setY(cursor.y() - 1);
	  if (cursor.x() > txt.lineLength (cursor.y()))
	       cursor.setX(txt.lineLength(cursor.y()));
	  break;
     case Key_Down:
	  if (cursor.y() < txt.getNumLines() - 1)
	       cursor.setY(cursor.y() + 1);
	  if (cursor.x() > txt.lineLength (cursor.y()))
	       cursor.setX(txt.lineLength(cursor.y()));
	  break;
     case Key_Left:
	  if (cursor.x() > 0)
	       cursor.setX(cursor.x() - 1);
	  else if (cursor.y() > 0) {
	       cursor.setY (cursor.y() - 1);
	       cursor.setX(txt.lineLength (cursor.y()));
	  }
	  break;
     case Key_Right:
	  if (cursor.x() < txt.lineLength(cursor.y()))
	       cursor.setX(cursor.x() + 1);
	  else if (cursor.y() < txt.getNumLines() - 1) {
	       cursor.setX(0);
	       cursor.setY (cursor.y() + 1);
	  }
	  break;
     case Key_Prior:
	  res = cursor.y() - height() / lh;
	  if (res < 0)
	       res = 0;
	  cursor.setY(res);
	  origin.setY(res);
      emit originChange (origin.y());
	  draw();
	  break;
     case Key_Next:
	  res = cursor.y() + height() / lh;
	  if (res >= txt.getNumLines())
	       res = txt.getNumLines() - 1;
	  cursor.setY(res);
	  origin.setY(res);
      emit originChange (origin.y());
	  draw();
	  break;
     case Key_Home:
          eraseCursor();
          cursor.setX (0);
          break;
     case Key_End:
          eraseCursor ();
          cursor.setX (txt.lineLength (cursor.y()));
          break;
     default:
	  if (e->ascii() < 32 || e->ascii() > 168)
	       switch (e->ascii()) {
	       case 1: // CTRL-A
		    eraseCursor();
		    cursor.setX (0);
		    break;
	       case 4: // CTRL-D
		    if (cursor.x() < txt.lineLength(cursor.y())) {
			 txt.deleteChar (cursor.y(), cursor.x());
			 drawLine();
			 changed = 1;
		    }
		    break;    
	       case 5: // CTRL-E
		    eraseCursor();
		    cursor.setX (txt.lineLength(cursor.y()));
		    break;
	       default:
                e->ignore();
	       }
	  else {
	       txt.insertChar (e->ascii(), cursor.y(), cursor.x());
	       cursor.setX(cursor.x()+1);
	       changed = 1;
	       drawLine();
	  }
     }
     if (cursor.y() > origin.y() - 2 + height() / lh) {
          if (cursor.y() < origin.y() + 1 + height () / lh)
               origin.setY(origin.y() + 3);
          else {
               unsigned int tmp = cursor.y() - height()/(lh << 1);
               if (tmp < 0)
                    tmp = 0;
               origin.setY(tmp);
          }
          emit originChange (origin.y());
          paintEvent (0);
     }
     if (cursor.y() < origin.y()) {
          if (cursor.y() > origin.y() - 3) 
               origin.setY(origin.y() - 3);
          else {
               unsigned int tmp = cursor.y() - height() / (lh << 1);
               if (tmp < 0)
                    tmp = 0;
               origin.setY(tmp);
          }
          emit originChange (origin.y());
          paintEvent(0);
     }
}

void DebugWidget::changeOrigin (int y)
{
     if (y < 0)
          y = 0;
     if (y >= txt.getNumLines ())
          y = txt.getNumLines ();
     origin.setY(y);
     draw();
}

QSize DebugWidget::sizeHint () const
{
     return QSize (100, 100);
}

QSizePolicy DebugWidget::sizePolicy () const
{
     return QSizePolicy (QSizePolicy::Expanding, QSizePolicy::Expanding);
}

void DebugWidget::resizeEvent (QResizeEvent *e)
{
     delete paint;
     delete pix;
     pix = new QPixmap (e->size().width(), e->size().height());
     paint = new QPainter (pix);
     emit originChange (origin.y());
}

void DebugWidget::mousePressEvent (QMouseEvent *e)
{
     eraseCursor();
     unsigned int line_nr = origin.y() + (e->y() - 10) / lh;
     if (line_nr < txt.getNumLines()) 
          if (e->x() < indent - SEP)
               if (line_nr == currentLine)
                    draggin = 1;
               else {
                    txt.reverseBreak(line_nr);
                    drawLine (line_nr);
               }
          else {
               unsigned short x = ((e->x() - indent) / metrics->width('x'));
               unsigned int y = ((e->y() - 10) / lh) + origin.y();
               cursor.setY (y >= txt.getNumLines() ? txt.getNumLines()-1 : y);
               cursor.setX (x <= txt.lineLength (cursor.y()) ?
                            x : txt.lineLength (cursor.y()));
               drawLine(line_nr);
               selecting = 1;
               selStart.setX(cursor.x());
               selStart.setY(cursor.y());
          }
}

void DebugWidget::mouseMoveEvent (QMouseEvent *e)
{
     if (draggin && e->x() < indent - SEP) {
          unsigned int line_nr = origin.y() + (e->y() - 10) / lh;
          if (line_nr != dragginLN) {
               cpaint->setRasterOp (Qt::CopyROP);
               cpaint->fillRect (LINENR,
                                 10 + (dragginLN - origin.y()) * lh,
                                 currentPix->width (),
                                 lh,
                                 lightGray);
               cpaint->drawImage (LINENR,
                                  10 + (line_nr - origin.y()) * lh +
                                  (lh - currentPix->height()) / 2,
                                  *currentPix );
               dragginLN = line_nr;
          }
     }
}

void DebugWidget::mouseReleaseEvent (QMouseEvent *e)
{
     unsigned int line_nr = origin.y() + (e->y() - 10) / lh;
     unsigned int old_line = currentLine;
     if (draggin && line_nr < txt.getNumLines ())
          if (e->x() < indent - SEP) {
               if (line_nr == currentLine)
                    txt.reverseBreak (line_nr);
               else {
                    currentLine = line_nr;
                    drawLine (old_line);
               }
               drawLine (currentLine);
          }
          else {
               selecting = 0;
               cursor.setY (line_nr);
               unsigned short x = e->x() / metrics->width ("x");
               cursor.setX ((x < txt.lineLength (line_nr) ? x :
                             txt.lineLength (line_nr)));
               drawLine (dragginLN);
               drawLine (currentLine);
          }
     draggin = 0;
}

void DebugWidget::focusInEvent (QFocusEvent *e)
{
     blinker->start(250);
}

void DebugWidget::focusOutEvent (QFocusEvent *e)
{
     eraseCursor();
     blinker->stop();
}

bool DebugWidget::focusNextPrevChild (bool next)
{
     return FALSE;
}

void DebugWidget::setFileName (const char *name)
{
     if (filename)
          delete filename;
     filename = new char[strlen(name) + 1];
     strcpy (filename, name);
}

void DebugWidget::loadFile ()
{
     loadFile (0);
}

void DebugWidget::loadFile (const char *fileName)
{
     if (!fileName) {
          QString tmp = QFileDialog::getOpenFileName("", "*.asm");
          if (!tmp.isNull()) {
               filename = new char [strlen (tmp.ascii()) + 1];
               strcpy (filename, tmp.ascii());
          }
          else {
               filename = 0;
               return;
          }
     }
     else {
          filename = new char [strlen (fileName) + 1];
          strcpy (filename, fileName);
     }
     txt.free();
     ifstream ifs (filename);
     int i = 0, j;
     char buf[151], c;
     c = ifs.get();
     while (!ifs.eof()) {
          for (j = 0; c != '\n' && j < 142; j++) {
               if (c == '\t') {
                    for (int k = 0; k < 8 && j < 150; k++)
                         buf[j++] = ' ';
                    j--;
               }
               else
                    buf[j] = c;
	       c = ifs.get();
          }
          buf[j] = 0;
          txt.insertChars(buf,i, 0);
          if (c == '\n') {
               i++;
	       c = ifs.get();
          }
     }
     changed = 1;
     cursor.setY(0);
     cursor.setX(0);
     origin.setX(0);
     origin.setY(0);
     currentLine = 0;
     draw();
     
     emit originChange (origin.y());
}

void DebugWidget::newFile ()
{
     origin.setY(0);
     origin.setX (0);
     cursor.setY(0);
     cursor.setX (0);
     currentLine = 0;
     txt.free ();
     txt.insertLine (0); 
     if (filename) {
          delete  [] filename;
          filename = 0;
     }
     emit originChange (origin.y());
     draw ();
}

void DebugWidget::saveFile ()
{
     saveFile (filename);
}

void DebugWidget::saveFile (const char *fileName)
{
     if (!fileName) {
          QString tmp = QFileDialog::getSaveFileName("", "*.asm");
          if (!tmp.isNull()) {
               filename = new char [strlen (tmp.ascii()) + 1];
               strcpy (filename, tmp.ascii());
          }
          else {
               filename = 0;
               return;
          }
     }
     else if (filename && filename != fileName) {
          delete [] filename;
          filename = new char [strlen (fileName) + 1];
          strcpy (filename, fileName);
     }
     cout << "Save file was here " << filename << endl;
     ofstream ofs (filename);
     for (int  i = 0; i < txt.getNumLines(); i++)
          ofs << txt.getChars (i) << endl;
     
}

int DebugWidget::curLineChanged()
{
     if (clchanged) {
          clchanged = 0;
          return 1;
     }
     return 0;
}

void DebugWidget::createLT (hashTable <unsigned int> *ht)
{
     ht->free (); // can I do this?
     
     line *tmp = txt.getLine (0);
     char buf[20];
     char *src, *dst;
     unsigned int ln = 0;
     while (tmp) {
          dst = buf;
          src = tmp->data;
          while (src && *src && *src != ' ' && *src != '\t' &&
                 *src != ';' && *src != ':')
               *dst = *src, dst++, src++;
          *dst = 0;
          if (buf[0]) {
               uppercase (buf);
               ht->add (buf, ln);
          }
          tmp = tmp->down;
          ln++;
     }
}
               
     
