/*
 * This file is part of the YASMIN 8051 simulator project
 *
 * Copyright (c) 2000 by Gwenn Englebienne
 *
 * This software is free software; it is protected by the GPL. You can 
 * modify and/or redistribute it under the terms specified by the GNU
 * Public Licence version 2.
 *
 * Be advised that this software comes with no warranty at all, not
 * even the implied warranty of merchantability or fitness for a 
 * particular purpose. 
 *
 * For more information, you could look at the COPYING file, or have a 
 * look at the free software foundation's web site: http://www.fsf.org
 *
 */
 
#include "memory.h"
#include <iostream.h>
#include <iomanip.h>

#define DEBUG 0

hashTable <SFR> sfr_info (20);

static struct {
     union {
          char direct[128];
          char indirect[128];
     } low;
     struct {
          char direct[128];
          char indirect[128];
     } high;
} RAM;

char prev_bank;

// Memory accessing functions

char& da (unsigned char index)
{
     if (index < 128)
          return RAM.low.direct[index];
     else 
          return RAM.high.direct[index-128];
}

char& ida (unsigned char index)
{
     if (index < 128)
          return RAM.low.indirect[index];
     else
          return RAM.high.indirect[index-128];
}

void set_bit (unsigned char address)
{
  unsigned char byte_addr;
  if (address & 0x80)  // it's an SFR
    byte_addr = address & 0xF8; // 11111000B
  else 
    byte_addr = (unsigned char)(short(address) >> 3) + 0x20;
  unsigned char bit_addr = address & 0x07;  // 00000111B
  char mask = 0x01 << bit_addr;
  da (byte_addr) |= mask;
}

void clr_bit (unsigned char address)
{
  unsigned char byte_addr;
  if (address & 0x80)  // it's an SFR
    byte_addr = address & 0xF8; // 11111000B
  else 
    byte_addr = (unsigned char)(short(address) >> 3) + 0x20;
  unsigned char bit_addr = address & 0x07;
  char mask = 0x01 << bit_addr;
  da (byte_addr) &= ~mask;
}

int is_bit_set (unsigned char address)
{
  unsigned char byte_addr;
  if (address & 0x80)  // it's an SFR
    byte_addr = address & 0xF8; // 11111000B
  else 
    byte_addr = (unsigned char)(short(address) >> 3) + 0x20;
  unsigned char bit_addr = address & 0x07;
  char mask = 0x01 << bit_addr;
  
  return (da (byte_addr) & mask) != 0;
}

void cpl_bit (unsigned char address)
{
  if (is_bit_set(address))
    clr_bit(address);
  else
    set_bit (address);
} 

unsigned short get_dptr (void)
{
  unsigned short tmp = (unsigned char)da (sfr_info["DPH"].address);
  tmp <<= 8;
  tmp |= (unsigned char)da (sfr_info["DPL"].address);
  return tmp;
}

void set_dptr (unsigned short val)
{
  unsigned short high = val;
  * (unsigned short *) &da (sfr_info["DPL"].address) = high;
  //  da (sfr_info["DPL"].address) = (val & 0xFF);
}

void create_sfr_with_bit_address (const char *name, unsigned char address)
{
  SFR tmp;
  tmp.type = byte;
  tmp.address = address;
  sfr_info[name] = tmp;

  char buf[10], *ptr = buf;
  const char *src = name;
  while (*src)
    *ptr++ = *src++;
  *ptr++ = '.';
  ptr++;
  *ptr = 0;
  ptr--;
  tmp.type = bit;
  for (int i = 0; i < 8; i++) {
    *ptr = '0' + i;
    tmp.address = address + i;
    if (DEBUG)
      cerr << "Adding bit address " << buf << " address = 0x" << setbase(16)
	   << (int) tmp.address << "\n";
    sfr_info[buf] = tmp;
  }
}

void create_r_registers (void)
{
//  if ((da (sfr_info["PSW"].address) & 0x18) != prev_bank) {
    int offset = da (sfr_info.get("PSW").address) & 0x18; // >> 3) * 8  !!
    if (DEBUG)
      cerr << "Registers offset == " << offset << ". PSW = "
	   << da (sfr_info.get("PSW").address) << endl;
    SFR tmp;
    tmp.type = byte;
    tmp.address = 0x00 + offset; sfr_info["R0"] = tmp;
    tmp.address = 0x01 + offset; sfr_info["R1"] = tmp;
    tmp.address = 0x02 + offset; sfr_info["R2"] = tmp;
    tmp.address = 0x03 + offset; sfr_info["R3"] = tmp;
    tmp.address = 0x04 + offset; sfr_info["R4"] = tmp;
    tmp.address = 0x05 + offset; sfr_info["R5"] = tmp;
    tmp.address = 0x06 + offset; sfr_info["R6"] = tmp;
    tmp.address = 0x07 + offset; sfr_info["R7"] = tmp;
    prev_bank = offset;
//  }
} 

void create_sfr_table (void)
{
     if (DEBUG)
	  cerr << "Creating sfr addresses table...";
     SFR tmp;
     tmp.type = byte;
     tmp.address = 0xE0; sfr_info["A"] = tmp;
     create_sfr_with_bit_address ("ACC", 0xE0);
     create_sfr_with_bit_address ("B", 0xF0);  
     create_sfr_with_bit_address ("PSW", 0xD0);
     tmp.address = 0x81; sfr_info["SP"] = tmp;
     tmp.address = 0x82; sfr_info["DPL"] = tmp;  
     tmp.address = 0x83; sfr_info["DPH"] = tmp;
     create_sfr_with_bit_address ("P0", 0x80);
     create_sfr_with_bit_address ("P1", 0x90);
     create_sfr_with_bit_address ("P2", 0xA0);
     create_sfr_with_bit_address ("P3", 0xB0);
     create_sfr_with_bit_address ("IP", 0xB8);
     create_sfr_with_bit_address ("IE", 0xA8);
     tmp.address = 0x89; sfr_info["TMOD"] = tmp;  
     create_sfr_with_bit_address ("TCON", 0x88);
     tmp.address = 0x8C; sfr_info["TH0"] = tmp;
     tmp.address = 0x8A; sfr_info["TL0"] = tmp;
     tmp.address = 0x8D; sfr_info["TH1"] = tmp;
     tmp.address = 0x8B; sfr_info["TL1"] = tmp;
     create_sfr_with_bit_address ("SCON", 0x98);
     tmp.address = 0x99; sfr_info["SBUF"] = tmp;
     tmp.address = 0x87; sfr_info["PCON"] = tmp;
     tmp.type = bit;
     tmp.address = 0xD7; sfr_info["C"] = tmp;
     tmp.address = 0xD5; sfr_info["F0"] = tmp;
     tmp.address = 0x88; sfr_info["IT0"] = tmp;
     tmp.address = 0x89; sfr_info["IE0"] = tmp;
     tmp.address = 0x8A; sfr_info["IT1"] = tmp;
     tmp.address = 0x8B; sfr_info["IE1"] = tmp;
     tmp.address = 0x8C; sfr_info["TR0"] = tmp;
     tmp.address = 0x8D; sfr_info["TF0"] = tmp;
     tmp.address = 0x8E; sfr_info["TR1"] = tmp;
     tmp.address = 0x8F; sfr_info["TF1"] = tmp;
     tmp.address = 0xA8; sfr_info["EX0"] = tmp;
     tmp.address = 0xA9; sfr_info["ET0"] = tmp;
     tmp.address = 0xAA; sfr_info["EX1"] = tmp;
     tmp.address = 0xAB; sfr_info["ET1"] = tmp;
     tmp.address = 0xAC; sfr_info["ES"] = tmp;
     tmp.address = 0xAD; sfr_info["ET2"] = tmp;
     tmp.address = 0xAE; sfr_info["EC"] = tmp;
     tmp.address = 0xAF; sfr_info["EA"] = tmp;
     tmp.address = 0xB8; sfr_info["PX0"] = tmp;
     tmp.address = 0xB9; sfr_info["PT0"] = tmp;
     tmp.address = 0xBA; sfr_info["PX1"] = tmp;
     tmp.address = 0xBB; sfr_info["PT1"] = tmp;
     tmp.address = 0xBC; sfr_info["PS"] = tmp;
     tmp.address = 0xBD; sfr_info["PT2"] = tmp;
     tmp.address = 0xBE; sfr_info["PPC"] = tmp;
     create_r_registers();
}

